/*

    Copyright (c) 2017 Motas Electronics Limited (J M HAYES)
    www.motas-synth.uk


    Permission is hereby granted, free of charge, to any person
    obtaining a copy of this software and any associated documentation files
    (the "Software"), to deal in the Software without restriction,
    including without limitation the rights to use, copy, modify, merge,
    publish, distribute, sublicense, and/or sell copies of the Software,
    and to permit persons to whom the Software is furnished to do so,
    subject to the following conditions:

    The above copyright notice and this permission notice shall be
    included in all copies or substantial portions of the Software.

    Any person wishing to distribute modifications to the Software is
    asked to send the modifications to the original developer so that
    they can be incorporated into the canonical version.  This is,
    however, not a binding provision of this license.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
    EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
    MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
    IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR
    ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
    CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
    WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.


 *
 * */

#include "utilities.h"

#include "QString"
#include "QFile"
#include "QDebug"
#include "QFileDialog"
#include <stdint.h>
#include "readscreen.h"
typedef union {
    struct {
        uint32_t ALL;
    };
    struct {
        uint8_t cin_cable;
        uint8_t evnt0;
        uint8_t evnt1;
        uint8_t evnt2;
    };
    struct {
        uint8_t type:4;
        uint8_t cable:4;
        uint8_t chn:4;
        uint8_t event:4;
        uint8_t value1;
        uint8_t value2;
    };
    struct {
        uint8_t cin:4;
        uint8_t dummy1_cable:4;
        uint8_t dummy1_chn:4;
        uint8_t dummy1_event:4;
        uint8_t note:8;
        uint8_t velocity:8;
    };
    struct {
        uint8_t dummy2_cin:4;
        uint8_t dummy2_cable:4;
        uint8_t dummy2_chn:4;
        uint8_t dummy2_event:4;
        uint8_t cc_number:8;
        uint8_t value:8;
    };
} midi_package_t;






Utilities::Utilities()
{
}


qint16 Utilities::convertMIDIbytes(quint8 currentByte, quint32 i)
{
    static quint8 msbByte = 0;
    static quint8 msbCount = 0;

    quint16 outVal = -1;
    if ((i % 8) == 0) //  this byte is the MSB for next 7 bytes, since cannot transmit values > 0x7F over MIDI
    {
        msbByte = currentByte;
        msbCount = 1;
    } else
    {
        outVal = currentByte | ((msbByte << msbCount) & 0x80);
        msbCount++;
    }
    return outVal;
}

quint32 Utilities::lowLevelCreateMIDIStream(quint8* buf, quint32 size,  quint8 type, quint8 startStopContinue,
        quint8 msg0, quint8 msg1, quint8 msg2, quint8* outBuf, quint8 model, quint8 channel)
{
    midi_package_t d;
    int32_t i = -4;

    quint16 remainder = size % 3;
    quint32 wholePackets = size / 3;
    quint32 packetsSent = 0;
    quint32 outCount = 0;
    while (1)
    {
     //   d.cin_cable = 0x04; // start/continue sysex
        if (i == -4)
        {
            d.evnt0 = 0xF0; // start sysex
            d.value1 = MIDI_stuff::BYTE_SYSEX_ID1;
            d.value2 = MIDI_stuff::BYTE_SYSEX_ID2;
            i++;
        }
        else if (i == -3)
        {
            d.evnt0 =  MIDI_stuff::BYTE_SYSEX_ID3;
            d.value1 = model;
            d.value2 = channel & 0x0F;
            i++;
        }
        else if (i == -2)
        {
            d.evnt0 = 0;
            d.value1 = type;
            d.value2 = startStopContinue;
            i++;
        }
        else if (i == -1 && startStopContinue != MIDI_SYSEX_START_TRANSFER && startStopContinue != MIDI_SYSEX_START_AND_END_TRANSFER)
        {
            i++;
            continue; // skip a transfer
        }
        else if (i == -1)
        {
            // other meta-data
            d.evnt0 = msg0;
            d.value1 = msg1;
            d.value2 = msg2;
            i++;
        }
        else if (packetsSent < wholePackets)
        {
            d.evnt0 = buf[i++];
            d.value1 = buf[i++];
            d.value2 = buf[i++];
            packetsSent++;
        } else // final byte
        {
            break;
        }

        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
        outBuf[outCount++] = d.value2;
    }

    // final packet
    if (remainder == 0)
    {
    //    d.cin_cable = 0x05; // end sysex
        d.evnt0 = 0xF7;
        d.value1 = 0;
        d.value2 = 0;
        outBuf[outCount++] = d.evnt0;
    }
    else if (remainder == 1)
    {
     //   d.cin_cable = 0x06; // end sysex
        d.evnt0 = buf[i] ;
        d.value1 = 0xF7;
        d.value2 = 0;
        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
    } else if (remainder == 2)
    {
      //  d.cin_cable = 0x07; // end sysex
        d.evnt0 = buf[i++] ;
        d.value1 = buf[i] ;
        d.value2 = 0xF7;
        outBuf[outCount++] = d.evnt0;
        outBuf[outCount++] = d.value1;
        outBuf[outCount++] = d.value2;
    }

    return outCount;
}




void Utilities::memcpyReverse32(char* out, char* in)
{
    for (int i = 0; i < 4; i++)
    {
        out[i] = in[i];
    }
}

void Utilities::memcpyForward32(char* out, char* in)
{
    for (int i = 0; i < 4; i++)
    {
        out[i] = in[3 - i];
    }
}



QVector<quint8> Utilities::prepareSysExPacket(int j, char* buffer, quint32 packetSize, quint32 fileLength, qint32 lastPacket, int type,
                                              quint8 msg0, quint8 msg1, quint8 msg2, quint8 modelNum, quint8 channel )
{
    QVector<quint8> message;
    message.append(MIDI_stuff::BYTE_SYSEX_START);
    message.append(MIDI_stuff::BYTE_SYSEX_ID1);
    message.append(MIDI_stuff::BYTE_SYSEX_ID2);

    message.append(MIDI_stuff::BYTE_SYSEX_ID3);
    message.append(modelNum);
    //message.append(MIDI_stuff::BYTE_SYSEX_PRODUCT_ID_2);
    message.append(channel & 0x0F);

    message.append(0);
    message.append(type);
    if (j == 0)
        message.append(MIDI_SYSEX_START_TRANSFER);
    else if (j == lastPacket - 1)
    {
        message.append(MIDI_SYSEX_END_TRANSFER);
        qDebug() << "MidiSettingsDialog:  FINAL firmware packet. " ;
    }
    else
        message.append(MIDI_SYSEX_CONTINUE_TRANSFER);


    if (j == 0)
    {
        message.append(msg0 & 0x7F);
        message.append(msg1 & 0x7F);
        message.append(msg2 & 0x7F);
    }
    for (quint32 i = 0 ; i < packetSize; i++)
    {
        quint32 index = i + j * packetSize;
        if (index >= fileLength)
            break;
        message.append(buffer[index]);
 //       firmwareBytesSent++;
    }


    message.append(MIDI_stuff::BYTE_SYSEX_END);
    return message;
}



quint32 Utilities::processBufferForMIDISend(char* outBuffer, char* buffer, quint32 originalDataLength)
{
    // now adjust binary file to allow send over MIDI - change data format to encode values > 0x7F
   // quint32 j = 0;
    quint32 k = 0;
    quint8 msb = 0;
    for (quint32 i = 0; i < originalDataLength; i += 7) // there is a MSB byte to be added for every 7 bytes, since cannot transmit values > 0x7F
    {
        msb = 0;
        for (int j = 0; j < 7; j++)
        {
            if ((i + j) >= originalDataLength)
                break;
            if ((quint8) buffer[i + j] > 0x7F)
                msb |= (0x40 >> j);
        }
        outBuffer[k++] = msb;
        for (int j = 0; j < 7; j++)
        {
            if ((i + j) >= originalDataLength)
                break;
            outBuffer[k++] = ((quint8) buffer[i + j]) & 0x7F;
        }
    }
    return k;
}


void Utilities::write32(quint32 val, int offset, quint8* buffer)
{

    for (int i = 0; i < 4; i++)
    {
         //qDebug() << (((quint32) this->midiStreamPatch->at(offset + i)) << (8 * (i)));
       buffer[offset + i]  = (val >> (8 * i)) & 0xFF;
    }
}

quint32 Utilities::read32(int offset, quint8* buffer)
{
    quint32 val = 0;
    for (int i = 0; i < 4; i++)
    {
         //qDebug() << (((quint32) this->midiStreamPatch->at(offset + i)) << (8 * (i)));
        val += ((quint32) buffer[offset + i]) << (8 * (i));
    }
    return val;
}


